/*
 * PROJECT: Indiana University Stiquito Colony Simulator and Control Software
 *
 * (c) Copyright 1996 Jason L. Almeter, Indiana University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that: (1) source distributions retain this entire copyright
 * notice and comment, and (2) distributions including binaries display
 * the following acknowledgement:  ``This product includes software
 * developed by Indiana University '' in the documentation or other
 * materials provided with the distribution and in all advertising
 * materials mentioning features or use of this software. The
 * name of the University may not be used to endorse or promote products
 * derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * MODULE:
 *	Robot Controllers
 *
 * FILE:
 *	robot.c
 *
 * ABSTRACT:
 *	avoids walls and robots
 *
 * $Revision: 1.4 $
 * $Date: 1996/01/19 14:29:18 $
 * $Author: jla $
 *
 * HISTORY:
 *
 * $Log: robot.c,v $
 * Revision 1.4  1996/01/19  14:29:18  jla
 * #ifdef'ed out unstable code for release.
 *
 * Revision 1.3  1996/01/17  19:17:06  jla
 * Updated Cogyright verbage for release.
 *
 * Revision 1.2  1996/01/16  20:04:35  jla
 * Added minimal access control to robots.
 *
 * Revision 1.1  1996/01/16  18:54:52  jla
 * Initial revision
 *
 * Revision 1.2  1996/01/15  14:49:26  jla
 * Enrolling in version control.
 *
 *
 */


#include <stdio.h>
#include <math.h>
#include "tca.h"
#include "../GPS/gps.h"
#include "../Motor/motor.h"
#include "../Clock/clock.h"
#include "robot.h"

#define DEBUG	1

#define MODULE	"Robot Module"

#define PI	3.14159265359

#define LARGE	(2.0*COLONY_SIZE)

/* our robot number */
int robotNumber;

CoordType loc;

int query;
int coords;
CoordType location[MAX_ROBOTS];

RobotCmdType command;

int time;		/* latest vitual time */

/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: void gpsHnd(TCA_REF_PTR ref, CoordArrayPtr c)
 *
 * DESCRIPTION:
 *	update robot locations with info from gps system
 *
\*--------------------------------------------------------------------------*/


void gpsHnd(TCA_REF_PTR ref, CoordArrayPtr c)
{
   int i;
   double theta;

   if(c->n > MAX_ROBOTS) {
      fprintf(stderr, "woah!!\n");
      exit(-1);
   }

   theta = location[robotNumber].theta;
   coords = 1;
   
   for(i=0; i<c->n; i++) {
      location[i].x = c->list[i].x;
      location[i].y = c->list[i].y;
      location[i].theta = c->list[i].theta;
   }

	/* if theta changed by roughly PI, must have gone in reverse */
   if(fabs(theta - location[robotNumber].theta) < 0.5*PI ||
      fabs(theta - location[robotNumber].theta) > 1.5*PI) {
   } else {
      location[robotNumber].theta += PI;
      if(location[robotNumber].theta < PI)
	 location[robotNumber].theta += 2.0*PI;
      if(location[robotNumber].theta > PI)
	 location[robotNumber].theta -= 2.0*PI;
   }

   tcaFreeData(tcaReferenceName(ref), c);
}

/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: void ToLocal(double x, double y, double theta,
 *		double *nx, double *ny)
 *
 * DESCRIPTION:
 *	convert x,y in colony coords into robot coords
 *
\*--------------------------------------------------------------------------*/


void ToLocal(double x, double y, double *nx, double *ny)
{
   *nx = (x*cos(-location[robotNumber].theta)
		- y*sin(-location[robotNumber].theta));
   *ny = (x*sin(-location[robotNumber].theta)
		+ y*cos(-location[robotNumber].theta));
}

/*--------------------------------------------------------------------------*\
 *
 * FUNCTION: void clockHnd(TCA_REF_PTR ref, ClockPtr t)
 *
 * DESCRIPTION:
 *	deal with updates from the vitual clock
 *
\*--------------------------------------------------------------------------*/


void clockHnd(TCA_REF_PTR ref, ClockPtr t)
{
   for(; time < *t; time++) {
      if(coords) {
	 robot_controller();
      }
   }
   tcaInform(RobotCmd, &command);/**/
   tcaFreeData(tcaReferenceName(ref), t);
}
   
main()
{
   int i,status;
   CookieT cookie;
   IdentT ident;
   char id[]="name";
   char ps[]="pass";


   query = 0;
   coords = 0;

   printf("Connect...\n");
   tcaConnectModule(MODULE, tcaServerMachine());

   tcaRegisterHandler(CoordBroadcast, "gpsHnd", gpsHnd);
   tcaLimitPendingMessages(CoordBroadcast, MODULE, 1);/**/

   tcaRegisterHandler(ClockBroadcast, "clockHnd", clockHnd);
   tcaLimitPendingMessages(ClockBroadcast, MODULE, 1);/**/
   
   tcaWaitUntilReady();

   ident.ident = id;
   ident.pass = ps;
   if(Success != tcaQuery(IncarnateQuery, &ident, &cookie)) {
      fprintf(stderr, "We were denied!\n");
      exit(-1);
   }

   robotNumber = cookie.number;

   printf("I am number %d %d\n", robotNumber, cookie.cookie);
   command.robot = robotNumber;
   command.cookie = cookie.cookie;

   /* this command kludges the simulator to put the robot in commission */
   command.cmd = CMD_F;

   tcaInform(RobotCmd, &command);/**/

   for(i=0; i<MAX_ROBOTS; i++) {
      location[i].x = LARGE;
      location[i].y = LARGE;
      location[i].theta = 0.5*PI;
   }

   tcaModuleListen();
}
