/*
 * PROJECT: Indiana University Stiquito Colony Simulator and Control Software
 *
 * (c) Copyright 1996 Jason L. Almeter, Indiana University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that: (1) source distributions retain this entire copyright
 * notice and comment, and (2) distributions including binaries display
 * the following acknowledgement:  ``This product includes software
 * developed by Indiana University '' in the documentation or other
 * materials provided with the distribution and in all advertising
 * materials mentioning features or use of this software. The
 * name of the University may not be used to endorse or promote products
 * derived from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * MODULE:
 *	Robot Controllers
 *
 * FILE:
 *	scent.c
 *
 * ABSTRACT:
 *	avoids walls and robots
 *	leaves a pheromone trail
 *
 * $Revision: 1.3 $
 * $Date: 1996/01/17 19:18:02 $
 * $Author: jla $
 *
 * HISTORY:
 *
 * $Log: scent.c,v $
 * Revision 1.3  1996/01/17  19:18:02  jla
 * Updated Cogyright verbage for release.
 *
 * Revision 1.2  1996/01/15  14:49:26  jla
 * Enrolling in version control.
 *
 *
 */


#include <stdio.h>
#include <math.h>
#include "tca.h"
#include "../GPS/gps.h"
#include "../Motor/motor.h"
#include "../Pheromone/pheromone.h"
#include "../Clock/clock.h"

#define DEBUG	1

#define MODULE	"Robot S Module"

#define PI	3.14159265359

#define FORWARD	1.0/12.0	/* +-FORWARD*PI goes forward */
#define F_TURN	4.0/6.0		/* these forward turn */
                                /* rest reverse turn */

#define QUIVER		0.1

#define WALL_ZONE	(ROBOT_RADIUS+ROBOT_RADIUS)
#define WALL_FORCE	1.0

#define ROBOT_ZONE	(3.0*ROBOT_RADIUS)
#define ROBOT_FORCE	2.0

#define LARGE	(2.0*COLONY_SIZE)

/* our robot number */
int number;

CoordType loc;

int query;
int coords;
CoordType robot[MAX_ROBOTS];

RobotCmdType command;

PheromoneT pheromone;

int time;

/* yummy new gps info */
void gpsHnd(TCA_REF_PTR ref, CoordArrayPtr c)
{
   int i;
   double theta;

   if(c->n > MAX_ROBOTS) {
      fprintf(stderr, "woah!!\n");
      exit(-1);
   }

   theta = robot[number].theta;
   coords = 1;
   
   for(i=0; i<c->n; i++) {
      robot[i].x = c->list[i].x;
      robot[i].y = c->list[i].y;
      robot[i].theta = c->list[i].theta;
   }

   if(fabs(theta - robot[number].theta) < 0.5*PI ||
      fabs(theta - robot[number].theta) > 1.5*PI) {
   } else {
      robot[number].theta += PI;
      if(robot[number].theta < PI)
	 robot[number].theta += 2.0*PI;
      if(robot[number].theta > PI)
	 robot[number].theta -= 2.0*PI;
   }

   tcaFreeData(tcaReferenceName(ref), c);
}

/* assume vector is relative to stiquito */
int vecToCommand(double x, double y)
{
   double theta;

   if(x == 0.0 && y == 0.0) {
      if(DEBUG) printf("%d stop\n",number);
      return CMD_STOP;
   }

/*   if(hypot(x,y) < QUIVER) {
/*      if(DEBUG) printf("inconclusive'\n");
/*      return CMD_STOP;
/*   }/**/
   
   theta = atan2(y,x);

/*   printf("%f\n",theta);/**/
   if(fabs(theta) < PI*FORWARD) {
      if(DEBUG) printf("%d go forward\n",number);
      return CMD_F;
   }

   if(theta < 0.0 && theta > -F_TURN*PI) {
      if(DEBUG) printf("%d go forward right\n",number);
      return CMD_FR;
   }
   
   if(theta > 0.0 && theta < F_TURN*PI) {
      if(DEBUG) printf("%d go forward left\n",number);
      return CMD_FL;
   }
   
   if(theta > F_TURN*PI) {
      if(DEBUG) printf("%d go reverse right\n",number);
      return CMD_RR;
   }
   
   if(theta < -F_TURN*PI) {
      if(DEBUG) printf("%d go reverse left\n",number);
      return CMD_RL;
   }

   fprintf(stderr, "%d what the hell?\n",number);
/*   exit(-1);/**/
   return CMD_STOP;
}

int foobar;

void controller()
{
   int i,forward;
   double dx,dy;
   double tx,ty;
   double norm;
   double scale;

   tx = 0.0;
   ty = 0.0;

   forward = 1;

   /* look for walls */
   if(fabs(NORTH_WALL - robot[number].y) < WALL_ZONE) {
      scale = (WALL_ZONE - fabs(NORTH_WALL - robot[number].y))/WALL_ZONE;
      ty -= WALL_FORCE*scale;
      forward = 0;
   }
   if(fabs(SOUTH_WALL - robot[number].y) < WALL_ZONE) {
      scale = (WALL_ZONE - fabs(SOUTH_WALL - robot[number].y))/WALL_ZONE;
      ty += WALL_FORCE*scale;
      forward = 0;
   }
   if(fabs(WEST_WALL - robot[number].x) < WALL_ZONE) {
      scale = (WALL_ZONE - fabs(WEST_WALL - robot[number].x))/WALL_ZONE;
      tx += WALL_FORCE*scale;
      forward = 0;
   }
   if(fabs(EAST_WALL - robot[number].x) < WALL_ZONE) {
      scale = (WALL_ZONE - fabs(EAST_WALL - robot[number].x))/WALL_ZONE;
      tx -= WALL_FORCE*scale;
      forward = 0;
   }
   /* convert to robot local coords */
   dx = (tx*cos(-robot[number].theta) - ty*sin(-robot[number].theta));
   dy = (tx*sin(-robot[number].theta) + ty*cos(-robot[number].theta));

   /* inter robot forces */
   for(i=0; i<MAX_ROBOTS; i++) {
      if(i != number) {
	 if(robot[i].x < EAST_WALL &&
	    robot[i].x > WEST_WALL &&
	    robot[i].y < NORTH_WALL &&
	    robot[i].y > SOUTH_WALL) {
	    tx = robot[i].x - robot[number].x;
	    ty = robot[i].y - robot[number].y;
	    norm = hypot(tx,ty);
	    if(norm < ROBOT_ZONE) {
	       scale = (ROBOT_ZONE - norm)/ROBOT_ZONE;
/*	       printf("too close %d %f\n",i,scale);/**/
	       tx /= norm;
	       ty /= norm;
	       dx += -scale*ROBOT_FORCE*(tx*cos(-robot[number].theta)
				   - ty*sin(-robot[number].theta));
	       dy += -scale*ROBOT_FORCE*(tx*sin(-robot[number].theta)
				   + ty*cos(-robot[number].theta));
	       forward = 0;
	    }
	 }
      }
   }

   /* forward */
   if(forward) {
      if(DEBUG) printf("...ROAM\n");
      dx += 2.0*QUIVER;
   }

/*   printf("%f\t%f\n",dx,dy);/**/
   command.cmd = vecToCommand(dx,dy);
/*   if(Success != tcaWaitForCommand(RobotCmd, &command)) {
      fprintf(stderr, "move failed!\n");
   }/**/
/*   tcaInform(RobotCmd, &command);/**/
   pheromone.next = 0;
   pheromone.x = robot[number].x;
   pheromone.y = robot[number].y;
   pheromone.color = 0;
   pheromone.expires = time;
   if(!foobar)
      tcaWaitForCommand(PheromoneCmd, &pheromone);
   foobar++;
   foobar %= 20;
}

void clockHnd(TCA_REF_PTR ref, ClockPtr t)
{
   for(; time < *t; time++) {
      if(coords) {
	 controller();
      }
   }
   if(foobar) {
      if(Success != tcaWaitForCommand(RobotCmd, &command)) {
	 fprintf(stderr, "move failed!\n");
      }/**/
   }
/*   printf(".");fflush(stdout);/**/
   tcaFreeData(tcaReferenceName(ref), t);
}
   
main()
{
   int i,status;

   query = 0;
   coords = 0;

   printf("Connect...\n");
   tcaConnectModule(MODULE, tcaServerMachine());

   tcaRegisterHandler(CoordBroadcast, "gpsHnd", gpsHnd);
   tcaLimitPendingMessages(CoordBroadcast, MODULE, 1);/**/

   tcaRegisterHandler(ClockBroadcast, "clockHnd", clockHnd);
   tcaLimitPendingMessages(ClockBroadcast, MODULE, 1);/**/
   
   tcaWaitUntilReady();

   tcaQuery(NumberQuery, NULL, &number);

   printf("I am number %d\n", number);
   command.robot = number;

   /* this command kludges the simulator to put the robot in commision */
   command.cmd = CMD_F;

/*   tcaInform(RobotCmd, &command);/**/
   if(Success != tcaWaitForCommand(RobotCmd, &command)) {
      fprintf(stderr, "move forward failed!\n");
   }/**/

   for(i=0; i<MAX_ROBOTS; i++) {
      robot[i].x = LARGE;
      robot[i].y = LARGE;
      robot[i].theta = 0.5*PI;
   }

   tcaModuleListen();

   for(;;) {
/*      status = tcaQuery(CoordQuery, &number, &loc);
/*         if(status == NullReply) {
/*   	 printf("no coords\n");
/*         } else {
/*   	 if(command.cmd == CMD_RR ||
/*   	    command.cmd == CMD_RL ||
/*   	    command.cmd == CMD_R) {
/*   	    robot[number].theta = robot[number].theta + PI;
/*   	 }
/*	 printf("gps: %f %f %f\n", robot[number].x, robot[number].y, robot[number].theta);/**/
/*         }
/*         if(query != number) {
/*   	 status = tcaQuery(CoordQuery, &query, &(robot[query]));
/*   	 if(status == NullReply) {
/*   	    robot[query].x = LARGE;
/*   	    robot[query].y = LARGE;
/*   	    printf("%d no coords\n",query);
/*   	 } else {
/*	    printf("gps %d: %f %f %f\n",
/*		   query,
/*		   robot[query].x,
/*		   robot[query].y,
/*		   robot[query].theta);/**/
/*   	 }
/*	 }/**/
      controller();
/*         if(++query >= MAX_ROBOTS)
/*   	 query -= MAX_ROBOTS;/**/
      while(Success == tcaHandleMessage(0));
      sleep(1);
   }
}
